"""Utilities for determining the settings available in nessai.

Used for bilby and pycbc-inference.
"""

from inspect import signature
from typing import List, Callable, Tuple


def _get_standard_methods() -> Tuple[List[Callable], List[Callable]]:
    """Get a list of the methods used by the standard sampler and the run
    method.
    """
    from ..flowsampler import FlowSampler
    from ..proposal import AugmentedFlowProposal, FlowProposal
    from ..samplers import NestedSampler

    methods = [
        AugmentedFlowProposal,
        FlowProposal,
        NestedSampler,
        FlowSampler,
    ]
    run_methods = [
        FlowSampler.run_standard_sampler,
    ]
    return methods, run_methods


def _get_importance_methods() -> Tuple[List[Callable], List[Callable]]:
    """Get a list of the methods used by the importance nested sampler and by
    the run method.
    """
    from ..flowsampler import FlowSampler
    from ..proposal.importance import ImportanceFlowProposal
    from ..samplers import ImportanceNestedSampler

    methods = [
        ImportanceFlowProposal,
        ImportanceNestedSampler,
        FlowSampler,
    ]
    run_methods = [
        FlowSampler.run_importance_nested_sampler,
    ]
    return methods, run_methods


def get_all_kwargs(
    importance_nested_sampler: bool = False,
    split_kwargs: bool = False,
) -> dict:
    """Get a dictionary of all possible kwargs and their default values.

    Parameters
    ----------
    importance_nested_sampler
        Indicates whether the importance nested sampler will be used or not.
        Defaults to :code:`False` for backwards compatibility.
    split_kwargs
        If True, the kwards are split into kwargs passed to :code:`FlowSampler`
        and those passes to :code:`FlowSampler.run`. If False, all kwargs are
        return in a single dictionary.

    Returns
    -------
    Dictionary of kwargs and their default values.
    """
    if importance_nested_sampler:
        methods, run_methods = _get_importance_methods()
    else:
        methods, run_methods = _get_standard_methods()

    kwargs = {}
    run_kwargs = {}
    for kwds, methods in zip([kwargs, run_kwargs], [methods, run_methods]):
        for m in methods:
            kwds.update(
                {
                    k: v.default
                    for k, v in signature(m).parameters.items()
                    if v.default is not v.empty
                }
            )

    if split_kwargs:
        return kwargs, run_kwargs
    else:
        kwargs.update(run_kwargs)
        return kwargs


def get_run_kwargs_list(importance_nested_sampler: bool = False) -> List[str]:
    """Get a list of kwargs used in the run method

    Parameters
    ----------
    importance_nested_sampler
        Indicates whether the importance nested sampler will be used or not.
        Defaults to :code:`False` for backwards compatibility.

    Returns
    -------
    List of kwargs.
    """
    from ..flowsampler import FlowSampler

    if importance_nested_sampler:
        method = FlowSampler.run_importance_nested_sampler
    else:
        method = FlowSampler.run_standard_sampler
    run_kwargs_list = [
        k
        for k, v in signature(method).parameters.items()
        if v.default is not v.empty
    ]
    return run_kwargs_list



"""Test the properties defined in ImportanceFlowProposal"""

from unittest.mock import patch
from nessai.proposal.importance import ImportanceFlowProposal as IFP
import numpy as np
import pytest


@pytest.fixture
def weights():
    return {"-1": 0.2, "0": 0.3, "1": 0.5}


@pytest.fixture
def ifp(ifp, weights):
    ifp._weights = weights
    return ifp


def test_weights(ifp, weights):
    assert IFP.weights.__get__(ifp) == weights


def test_weights_array(ifp, weights):
    np.testing.assert_array_equal(
        IFP.weights_array.__get__(ifp), np.array(list(weights.values()))
    )


def test_n_proposals(ifp):
    IFP.n_proposals.__get__(ifp) == 3


def test_flow_config(ifp):
    flow_config = {"a": 1}
    ifp._flow_config = flow_config
    IFP.flow_config.__get__(ifp) is flow_config


@pytest.mark.parametrize("config", [None, {}, {"model_config": {}}])
def test_flow_config_setter(ifp, config):
    ifp.model.dims = 2
    out = {"test": True, "model_config": {"n_inputs": 2}}
    with patch(
        "nessai.proposal.importance.update_config", return_value=out
    ) as mock:
        IFP.flow_config.__set__(ifp, config)
    mock.assert_called_once_with({"model_config": {"n_inputs": 2}})
    assert ifp._flow_config is out


@pytest.mark.parametrize(
    "reset_flow, level_count",
    [[1, 3], [4, 8], [True, 5]],
)
def test_reset_flow_true(ifp, reset_flow, level_count):
    ifp.reset_flow = reset_flow
    ifp.level_count = level_count
    assert IFP._reset_flow.__get__(ifp) is True


@pytest.mark.parametrize(
    "reset_flow, level_count",
    [[False, 4], [5, 8]],
)
def test_reset_flow_false(ifp, reset_flow, level_count):
    ifp.reset_flow = reset_flow
    ifp.level_count = level_count
    assert IFP._reset_flow.__get__(ifp) is False



# -*- coding: utf-8 -*-
"""
General tests of `get_reparameterisation`.
"""
import pytest

from nessai.reparameterisations import (
    Angle,
    AnglePair,
    NullReparameterisation,
    Reparameterisation,
    Rescale,
    RescaleToBounds,
    ToCartesian,
    get_reparameterisation,
)

# List of known reparameterisations, the class and expected kwargs.
known_reparameteristions = [
    ("default", RescaleToBounds, {}),
    ("rescaletobounds", RescaleToBounds, {}),
    ("rescale-to-bounds", RescaleToBounds, {}),
    ("offset", RescaleToBounds, {"offset": True}),
    (
        "inversion",
        RescaleToBounds,
        {
            "detect_edges": True,
            "boundary_inversion": True,
            "inversion_type": "split",
        },
    ),
    (
        "inversion-duplicate",
        RescaleToBounds,
        {
            "detect_edges": True,
            "boundary_inversion": True,
            "inversion_type": "duplicate",
        },
    ),
    (
        "logit",
        RescaleToBounds,
        {
            "update_bounds": False,
            "rescale_bounds": [0.0, 1.0],
            "post_rescaling": "logit",
        },
    ),
    (
        "log-rescale",
        RescaleToBounds,
        {
            "rescale_bounds": [0.0, 1.0],
            "update_bounds": False,
            "post_rescaling": "log",
        },
    ),
    ("scale", Rescale, {}),
    ("rescale", Rescale, {}),
    ("angle", Angle, {}),
    ("angle-pi", Angle, {"scale": 2.0, "prior": "uniform"}),
    ("angle-2pi", Angle, {"scale": 1.0, "prior": "uniform"}),
    ("periodic", Angle, {"scale": None}),
    ("angle-sine", RescaleToBounds, {}),
    ("angle-cosine", RescaleToBounds, {}),
    ("angle-pair", AnglePair, {}),
    ("to-cartesian", ToCartesian, {}),
    ("none", NullReparameterisation, {}),
    ("null", NullReparameterisation, {}),
    (None, NullReparameterisation, {}),
]


@pytest.mark.parametrize(
    "inputs",
    known_reparameteristions,
)
def test_get_reparameterisation(inputs):
    """Test all of the included reparameterisations."""
    name, expected_class, expected_kwargs = inputs
    reparam, kwargs = get_reparameterisation(name)
    assert reparam is expected_class
    assert kwargs == expected_kwargs


def test_get_reparameterisation_with_class():
    """Test the case of class that inherits from Reparameterisation"""

    class TestReparam(Reparameterisation):
        pass

    reparam, kwargs = get_reparameterisation(TestReparam)
    assert reparam is TestReparam
    assert not kwargs


def test_get_reparameterisation_with_class_error():
    """Test the case of class that doe not inherit from Reparameterisation.

    This should raise an error.
    """

    class TestReparam:
        pass

    with pytest.raises(TypeError) as excinfo:
        get_reparameterisation(TestReparam)
    assert "Reparameterisation must be" in str(excinfo.value)


def test_get_reparameterisation_unknown_name():
    """Assert an error is raised in the string is not recognised."""
    with pytest.raises(ValueError) as excinfo:
        get_reparameterisation("shift")
    assert "Unknown reparameterisation: shift" in str(excinfo.value)


def test_get_reparameterisation_invalid_input():
    """Assert an error is raised if the input type is invalid."""
    with pytest.raises(TypeError) as excinfo:
        get_reparameterisation(2)
    assert "Reparameterisation must be" in str(excinfo.value)


def test_get_reparameterisation_defaults():
    """Test using an updated defaults dictionary."""
    defaults = {"default": ("Class", {"x": 2})}
    cls, kwargs = get_reparameterisation("default", defaults=defaults)
    assert cls == "Class"
    assert kwargs == {"x": 2}



import datetime
import numpy as np
import pickle
import pytest
from unittest.mock import MagicMock, patch

from nessai.samplers.importancesampler import ImportanceNestedSampler as INS


def test_getstate_no_model(ins):
    ins.proposal = MagicMock()
    ins.model = None
    state, proposal, training_samples, iid_samples = INS.__getstate__(ins)
    assert "model" not in state
    assert state["_previous_likelihood_evaluations"] == 0
    assert state["_previous_likelihood_evaluation_time"] == 0
    assert proposal is ins.proposal
    assert training_samples is ins.training_samples
    assert iid_samples is ins.iid_samples


def test_getstate_model(ins):

    evals = 10
    time = datetime.timedelta(seconds=30)

    ins.proposal = MagicMock()
    ins.model = MagicMock()
    ins.model.likelihood_evaluations = evals
    ins.model.likelihood_evaluation_time = time

    state, proposal, training_samples, iid_samples = INS.__getstate__(ins)
    assert "model" not in state
    assert state["_previous_likelihood_evaluations"] == evals
    assert state["_previous_likelihood_evaluation_time"] == 30
    assert proposal is ins.proposal
    assert training_samples is ins.training_samples
    assert iid_samples is ins.iid_samples


@pytest.mark.parametrize("has_log_q", [False, True])
def test_resume_from_pickled_sampler(model, samples, has_log_q):

    sampler = MagicMock()

    obj = MagicMock()
    obj.log_q = None
    obj.log_evidence = 0.0
    obj.log_evidence_error = 1.0
    obj.proposal = MagicMock()
    log_meta_proposal = np.log(np.random.rand(len(samples)))
    log_q = np.log(np.random.rand(len(samples)))
    log_meta_proposal_iid = np.log(np.random.rand(len(samples)))
    log_q_iid = np.log(np.random.rand(len(samples)))
    obj.proposal.compute_meta_proposal_samples = MagicMock(
        side_effect=[
            (log_meta_proposal, log_q),
            (log_meta_proposal_iid, log_q_iid),
        ]
    )
    obj.training_samples.samples = samples
    obj.iid_samples.samples = samples
    if has_log_q:
        obj.training_samples.log_q = log_q
        obj.iid_samples.log_q = log_q_iid
    else:
        obj.training_samples.log_q = None
        obj.iid_samples.log_q = None

    with patch(
        "nessai.samplers.importancesampler.BaseNestedSampler.resume_from_pickled_sampler",  # noqa
        return_value=obj,
    ) as mock_resume:
        out = INS.resume_from_pickled_sampler(sampler, model)

    mock_resume.assert_called_once_with(sampler, model)
    if has_log_q:
        obj.proposal.compute_meta_proposal_samples.assert_not_called()
    else:
        obj.proposal.compute_meta_proposal_samples.assert_called()

    assert out.training_samples.log_q is log_q
    assert out.iid_samples.log_q is log_q_iid


@pytest.mark.parametrize("save_log_q", [True, False])
@pytest.mark.integration_test
def test_pickling_sampler_integration(integration_model, tmp_path, save_log_q):
    outdir = tmp_path / "test_pickle"
    ins = INS(
        model=integration_model,
        output=outdir,
        nlive=50,
        min_samples=10,
        max_iteration=1,
        save_log_q=save_log_q,
        plot=False,
        checkpointing=False,
    )
    ins.nested_sampling_loop()
    data = pickle.dumps(ins)
    loaded_ins = pickle.loads(data)
    if save_log_q:
        np.testing.assert_array_equal(
            loaded_ins._ordered_samples.log_q, ins._ordered_samples.log_q
        )
    else:
        assert loaded_ins._ordered_samples.log_q is None



# -*- coding: utf-8 -*-
"""Test the augment proposal"""
import numpy as np
import pytest
from unittest.mock import create_autospec, MagicMock, patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.proposal import AugmentedFlowProposal


@pytest.fixture
def proposal():
    return create_autospec(AugmentedFlowProposal)


@pytest.fixture
def x():
    _x = np.concatenate(
        [np.random.rand(10, 2), np.random.randn(10, 2)], axis=1
    )
    return numpy_array_to_live_points(_x, ["x", "y", "e_1", "e_2"])


def test_init(model):
    """Test the init function"""
    AugmentedFlowProposal(model, poolsize=100)


def test_update_flow_config(proposal):
    """Test update flow config"""
    proposal.rescaled_dims = 4
    proposal.augment_dims = 2
    proposal.flow_config = dict(model_config={})
    with patch(
        "nessai.proposal.augmented.FlowProposal.update_flow_config"
    ) as mock:
        AugmentedFlowProposal.update_flow_config(proposal)
    mock.assert_called_once()
    mask = np.array([1, 1, -1, -1])
    np.testing.assert_array_equal(
        proposal.flow_config["model_config"]["kwargs"]["mask"], mask
    )


@patch("nessai.proposal.FlowProposal.set_rescaling")
def test_set_rescaling(mock, proposal):
    """Test the set rescaling method"""
    proposal.names = ["x", "y"]
    proposal.rescaled_names = ["x_prime", "y_prime"]
    proposal.rescale_parameters = proposal.names
    proposal.augment_dims = 2
    AugmentedFlowProposal.set_rescaling(proposal)

    assert proposal.names == ["x", "y", "e_0", "e_1"]
    assert proposal.rescaled_names == ["x_prime", "y_prime", "e_0", "e_1"]
    assert proposal.augment_names == ["e_0", "e_1"]
    mock.assert_called_once()


@pytest.mark.parametrize("generate", ["zeroes", "gaussian"])
@patch("numpy.random.randn", return_value=np.ones(10))
@patch("numpy.zeros", return_value=np.ones(10))
def test_rescaling(mock_zeros, mock_randn, proposal, x, generate):
    """Test the rescaling method"""
    proposal._base_rescale = MagicMock(return_value=[x, np.ones(x.size)])
    proposal.augment_names = ["e_1"]
    proposal.augment_dims = 1

    AugmentedFlowProposal._augmented_rescale(
        proposal, x, generate_augment=generate, test=True
    )

    proposal._base_rescale.assert_called_once_with(
        x, compute_radius=False, test=True
    )

    if generate == "zeroes":
        mock_zeros.assert_called_once_with(x.size)
    else:
        mock_randn.assert_called_once_with(x.size)


@pytest.mark.parametrize("compute_radius", [False, True])
@patch("numpy.random.randn", return_value=np.arange(10))
@patch("numpy.zeros", return_value=np.ones(10))
def test_rescaling_generate_none(
    mock_zeros, mock_randn, proposal, x, compute_radius
):
    """Test the rescaling method with generate_augment=None"""
    proposal._base_rescale = MagicMock(return_value=[x, np.ones(x.size)])
    proposal.augment_names = ["e_1"]
    proposal.augment_dims = 1
    proposal.generate_augment = "zeros"

    AugmentedFlowProposal._augmented_rescale(
        proposal,
        x,
        generate_augment=None,
        test=True,
        compute_radius=compute_radius,
    )

    proposal._base_rescale.assert_called_once_with(
        x, compute_radius=compute_radius, test=True
    )

    if not compute_radius:
        mock_randn.assert_called_once_with(x.size)
    else:
        mock_zeros.assert_called_once_with(x.size)


def test_rescaling_generate_unknown(proposal, x):
    """Test the rescaling method with with an unknown method for generate"""
    proposal._base_rescale = MagicMock(return_value=[x, np.ones(x.size)])

    with pytest.raises(RuntimeError) as excinfo:
        AugmentedFlowProposal._augmented_rescale(
            proposal, x, generate_augment="ones"
        )
    assert "Unknown method" in str(excinfo.value)


@pytest.mark.parametrize("marg", [False, True])
@patch("scipy.stats.norm.logpdf")
def test_augmented_prior(mock, marg, proposal, x):
    """Test the augmented prior with and without marginalistion"""
    proposal.marginalise_augment = marg
    proposal.augment_names = ["e_1", "e_2"]
    log_prior = AugmentedFlowProposal.augmented_prior(proposal, x)
    if marg:
        assert log_prior == 0
    else:
        np.testing.assert_array_equal(x["e_1"], mock.call_args_list[0][0][0])
        np.testing.assert_array_equal(x["e_2"], mock.call_args_list[1][0][0])


@patch("nessai.proposal.flowproposal.FlowProposal.log_prior", return_value=1)
def test_log_prior(mock_prior, proposal, x):
    """Test the complete log prior"""
    proposal.augmented_prior = MagicMock(return_value=1)

    log_p = AugmentedFlowProposal.log_prior(proposal, x)

    mock_prior.assert_called_once_with(x)
    proposal.augmented_prior.assert_called_once_with(x)
    assert log_p == 2


@patch(
    "nessai.proposal.flowproposal.FlowProposal.x_prime_log_prior",
    return_value=1,
)
def test_prime_log_prior(mock_prior, proposal, x):
    """Test the complete prime log prior"""
    proposal.augmented_prior = MagicMock(return_value=1)

    log_p = AugmentedFlowProposal.x_prime_log_prior(proposal, x)

    mock_prior.assert_called_once_with(x)
    proposal.augmented_prior.assert_called_once_with(x)
    assert log_p == 2


def test_marginalise_augment(proposal):
    """Test the marginalise augment function"""
    proposal.n_marg = 5
    proposal.augment_dims = 2
    x = np.concatenate([np.random.randn(3, 2), np.zeros((3, 2))], axis=1)
    z = np.random.randn(15, 4)
    log_prob = np.random.randn(15)
    proposal.flow = MagicMock()
    proposal.flow.forward_and_log_prob = MagicMock(return_value=[z, log_prob])
    log_prob_out = AugmentedFlowProposal._marginalise_augment(proposal, x)

    assert len(log_prob_out) == 3


@pytest.mark.parametrize("log_p", [np.ones(2), np.array([-1, np.inf])])
@pytest.mark.parametrize("marg", [False, True])
def test_backward_pass(proposal, model, log_p, marg):
    """Test the backward pass method"""
    n = 2
    acc = int(np.isfinite(log_p).sum())
    x = np.random.randn(n, model.dims)
    z = np.random.randn(n, model.dims)
    proposal._marginalise_augment = MagicMock(return_value=log_p)
    proposal.inverse_rescale = MagicMock(
        side_effect=lambda a: (a, np.ones(a.size))
    )
    proposal.rescaled_names = model.names
    proposal.alt_dist = None
    proposal.check_prior_bounds = MagicMock(side_effect=lambda a, b: (a, b))
    proposal.flow = MagicMock()
    proposal.flow.sample_and_log_prob = MagicMock(return_value=[x, log_p])

    proposal.marginalise_augment = marg

    x_out, log_p = AugmentedFlowProposal.backward_pass(proposal, z)

    assert len(x_out) == acc
    proposal.inverse_rescale.assert_called_once()
    proposal.flow.sample_and_log_prob.assert_called_once_with(
        z=z, alt_dist=None
    )

    assert proposal._marginalise_augment.called is marg


@pytest.mark.integration_test
def test_w_default_rescaling(model, tmpdir):
    """Integration test to make sure augmented proposal works with the default
    rescaling method.
    """
    x = model.new_point(100)

    x_out = (
        2 * (x["x"] - model.bounds["x"][0]) / (np.ptp(model.bounds["x"])) - 1
    )
    y_out = (
        2 * (x["y"] - model.bounds["y"][0]) / (np.ptp(model.bounds["y"])) - 1
    )

    output = tmpdir.mkdir("testdir")
    proposal = AugmentedFlowProposal(
        model,
        output=output,
        poolsize=100,
        augment_dims=2,
        rescale_parameters=True,
        update_bounds=False,
    )

    proposal.initialise()

    assert proposal.rescaled_names == ["x_prime", "y_prime", "e_0", "e_1"]

    x_prime, log_j = proposal.rescale(x)

    np.testing.assert_array_equal(x_out, x_prime["x_prime"])
    np.testing.assert_array_equal(y_out, x_prime["y_prime"])

    x_inv, log_j_inv = proposal.inverse_rescale(x_prime)

    np.testing.assert_array_equal(log_j, -log_j_inv)
    np.testing.assert_array_equal(x[["x", "y"]], x_inv[["x", "y"]])


@pytest.mark.integration_test
def test_w_reparameterisation(model, tmpdir):
    """Integration test to make sure augmented proposal works with
    reparameterisaitons configured using the dictionary method.
    """
    x = model.new_point(100)
    reparameterisations = {
        "x": {"reparameterisation": "rescaletobounds", "update_bounds": False},
        "y": {"reparameterisation": "scale", "scale": 2.0},
    }

    x_out = (
        2 * (x["x"] - model.bounds["x"][0]) / (np.ptp(model.bounds["x"])) - 1
    )
    y_out = x["y"] / 2.0

    output = tmpdir.mkdir("testdir")
    proposal = AugmentedFlowProposal(
        model,
        output=output,
        poolsize=100,
        augment_dims=2,
        reparameterisations=reparameterisations,
    )

    proposal.initialise()

    assert proposal.rescaled_names == ["x_prime", "y_prime", "e_0", "e_1"]

    x_prime, log_j = proposal.rescale(x)

    np.testing.assert_array_equal(x_out, x_prime["x_prime"])
    np.testing.assert_array_equal(y_out, x_prime["y_prime"])

    x_inv, log_j_inv = proposal.inverse_rescale(x_prime)

    np.testing.assert_array_equal(log_j, -log_j_inv)
    np.testing.assert_array_equal(x[["x", "y"]], x_inv[["x", "y"]])



# -*- coding: utf-8 -*-
"""
Proposal method for initial sampling when priors are not analytical.
"""
import numpy as np

from .analytic import AnalyticProposal


class RejectionProposal(AnalyticProposal):
    """Object for rejection sampling from the priors.

    See parent for explanation of arguments and keyword arguments.

    Will be used when ``nessai`` is called with ``analytic_priors=False``. This
    is the default behaviour.

    Relies on :py:meth:`nessai.model.Model.new_point` to draw new points and
    :py:meth:`nessai.model.Model.new_point_log_prob` when computing the
    probability of each new point.
    """

    def __init__(self, *args, **kwargs):
        super(RejectionProposal, self).__init__(*args, **kwargs)
        self._checked_population = True
        self.population_acceptance = None

    def draw_proposal(self, N=None):
        """Draw new point(s).

        Parameters
        ----------
        N : int, optional
            Number of samples to draw. If not specified ``poolsize`` will be
            used.

        Returns
        -------
        structured_array
            Array of N new points
        """
        if N is None:
            N = self.poolsize
        return self.model.new_point(N=N)

    def log_proposal(self, x):
        """
        Log proposal probability. Calls \
                :meth:`nessai.model.Model.new_point_log_prob`

        Parameters
        ----------
        x : structured_array
            Array of new points

        Returns
        -------
        :obj:`numpy.ndarray`
            Array of log-probabilities.
        """
        return self.model.new_point_log_prob(x)

    def compute_weights(self, x, return_log_prior=False):
        """
        Get weights for the samples.

        Computes the log weights for rejection sampling sampling but does not
        normalize the weights.

        Parameters
        ----------
        x :  structured_array
            Array of points
        return_log_prior: bool
            If true, the log-prior probability is also returned.

        Returns
        -------
        log_w : :obj:`numpy.ndarray`
            Array of log-weights rescaled such that the maximum value is zero.
        """
        log_p = self.model.batch_evaluate_log_prior(x)
        log_q = self.log_proposal(x)
        log_w = log_p - log_q
        if return_log_prior:
            return log_w, log_p
        else:
            return log_w

    def populate(self, N=None):
        """
        Populate the pool by drawing from the proposal distribution and
        using rejection sampling.

        Will also evaluate the likelihoods if the proposal contains a
        multiprocessing pool.

        Parameters
        ----------
        N : int, optional
            Number of samples to draw. Not all samples will be accepted to
            the number of samples saved will be less than N. If not specified
            ``poolsize`` will be used.
        """
        if N is None:
            N = self.poolsize
        x = self.draw_proposal(N=N)
        log_w, x["logP"] = self.compute_weights(x, return_log_prior=True)
        log_w -= np.nanmax(log_w)
        log_u = np.log(np.random.rand(N))
        indices = np.where((log_w - log_u) >= 0)[0]
        self.samples = x[indices]
        self.indices = np.random.permutation(self.samples.shape[0]).tolist()
        self.population_acceptance = self.samples.size / N
        self.samples["logL"] = self.model.batch_evaluate_log_likelihood(
            self.samples
        )
        self.populated = True
        self._checked_population = False



#!/usr/bin/env python

# Example of using Nessai with Bilby (Requires separate installation)
# with unbounded priors.

import bilby
import numpy as np

outdir = "./outdir/"
label = "bilby_unbounded_priors"

# Setup the bilby logger
bilby.core.utils.setup_logger(outdir=outdir, label=label)

# Define a likelihood using Bilby


class SimpleGaussianLikelihood(bilby.Likelihood):
    def __init__(self):
        """A very simple Gaussian likelihood"""
        super().__init__(parameters={"x": None, "y": None})

    def log_likelihood(self):
        """Log-likelihood."""
        return -0.5 * (
            self.parameters["x"] ** 2.0 + self.parameters["y"] ** 2.0
        ) - np.log(2.0 * np.pi)


# Define priors, we'll use Gaussians since they're unbounded.
priors = dict(
    x=bilby.core.prior.Gaussian(0, 5, "x"),
    y=bilby.core.prior.Gaussian(0, 10, "y"),
)

# Instantiate the likelihood
likelihood = SimpleGaussianLikelihood()

# Run using bilby.run_sampler, any kwargs are parsed to the sampler
# NOTE: when using Bilby if the priors can be sampled analytically  the flag
# `analytic_priors` enables faster initial sampling. See 'further details' in
# the documentation for more details
# We need to disable the rescaling since we can't rescale without bounds.
# Alternatively a different reparameterisation could be used, in this case
# we can used the 'Rescale' reparameterisation that rescales the inputs by
# a constant, this is useful when we do not have prior bounds we can use.
result = bilby.run_sampler(
    outdir=outdir,
    label=label,
    resume=False,
    plot=True,
    likelihood=likelihood,
    priors=priors,
    sampler="nessai",
    injection_parameters={"x": 0.0, "y": 0.0},
    reparameterisations={
        "scale": {"parameters": ["x", "y"], "scale": [5, 10]}
    },
    analytic_priors=True,
    seed=1234,
)



"""Tests related to how samples are handled"""

from unittest.mock import MagicMock, create_autospec

from nessai.livepoint import numpy_array_to_live_points
from nessai.samplers.importancesampler import (
    ImportanceNestedSampler as INS,
    OrderedSamples,
)
import numpy as np


def test_ordered_samples_property(ins):
    ins.training_samples = object()
    ins.draw_iid_live = False
    assert INS._ordered_samples.__get__(ins) is ins.training_samples


def test_ordered_samples_property_iid(ins):
    ins.iid_samples = object()
    ins.draw_iid_live = True
    assert INS._ordered_samples.__get__(ins) is ins.iid_samples


def test_live_points_unit_property(ins):
    ins._ordered_samples = MagicMock(spec=OrderedSamples)
    assert (
        INS.live_points_unit.__get__(ins) is ins._ordered_samples.live_points
    )


def test_nested_samples_unit_property(ins):
    ins._ordered_samples = MagicMock(spec=OrderedSamples)
    assert (
        INS.nested_samples_unit.__get__(ins)
        is ins._ordered_samples.nested_samples
    )


def test_nested_samples_property(ins):
    ins._ordered_samples = MagicMock(spec=OrderedSamples)
    assert (
        INS.nested_samples_unit.__get__(ins)
        is ins._ordered_samples.nested_samples
    )


def test_populate_live_points_no_iid(ins, model):
    n = 100
    ins.n_initial = n
    ins.model = model
    ins.sample_counts = {}
    ins.draw_iid_live = False

    INS.populate_live_points(ins)

    ins.training_samples.add_initial_samples.assert_called_once()
    assert len(ins.training_samples.add_initial_samples.call_args.args[0]) == n
    assert ins.training_samples.add_initial_samples.call_args.args[
        1
    ].shape == (n, 1)


def test_populate_live_points_iid(ins, model):
    n = 100
    ins.n_initial = n
    ins.model = model
    ins.draw_iid_live = True
    ins.sample_counts = {}
    ins.iid_samples = create_autospec(OrderedSamples)

    INS.populate_live_points(ins)

    ins.training_samples.add_initial_samples.assert_called_once()
    assert len(ins.training_samples.add_initial_samples.call_args.args[0]) == n
    assert ins.training_samples.add_initial_samples.call_args.args[
        1
    ].shape == (n, 1)
    assert np.isfinite(
        ins.training_samples.add_initial_samples.call_args.args[0]["logL"]
    ).all()
    assert np.isfinite(
        ins.training_samples.add_initial_samples.call_args.args[0]["logP"]
    ).all()

    ins.iid_samples.add_initial_samples.assert_called_once()
    assert len(ins.iid_samples.add_initial_samples.call_args.args[0]) == n
    assert ins.iid_samples.add_initial_samples.call_args.args[1].shape == (
        n,
        1,
    )
    assert np.isfinite(
        ins.iid_samples.add_initial_samples.call_args.args[0]["logL"]
    ).all()
    assert np.isfinite(
        ins.iid_samples.add_initial_samples.call_args.args[0]["logP"]
    ).all()


def test_remove_samples(ins, iid):
    ins.history = {}
    ins.history["n_removed"] = [3]
    ins.training_samples.remove_samples = MagicMock(return_value=5)
    ins.draw_iid_live = iid
    expected = 5
    if iid:
        ins.iid_samples = create_autospec(OrderedSamples)
        ins.iid_samples.remove_samples = MagicMock(return_value=4)
        expected = 4

    out = INS.remove_samples(ins)

    ins.training_samples.remove_samples.assert_called_once()
    assert out == expected
    assert ins.history["n_removed"] == [3, expected]


def test_adjust_final_samples(ins, proposal, model, samples, log_q):
    def draw(n, flow_number=None, update_counts=False):
        assert update_counts is False
        x = numpy_array_to_live_points(
            np.random.randn(n, model.dims),
            names=model.names,
        )
        lq = np.random.rand(n, log_q.shape[1])
        return x, lq

    def draw_from_prior(n):
        x = model.new_point(n)
        lq = np.random.rand(n, log_q.shape[1])
        return x, lq

    proposal.draw = MagicMock(side_effect=draw)
    proposal.draw_from_prior = MagicMock(side_effect=draw_from_prior)
    proposal.n_requested = {"-1": 10, "1": 10}

    ins.samples_unit = samples
    ins.log_q = log_q
    ins.proposal = proposal
    ins.model = model

    INS.adjust_final_samples(ins)


def test_compute_importance(ins, iid):
    importance = {"evidence": 0.4, "posterior": 0.8}
    ins.training_samples = MagicMock(spec=OrderedSamples)
    ins.training_samples.compute_importance = MagicMock(
        return_value=importance
    )
    ins.iid_samples = MagicMock(spec=OrderedSamples)
    ins.iid_samples.compute_importance = MagicMock(return_value=importance)
    ins.draw_iid_live = iid

    out = INS.compute_importance(ins)
    assert out is importance
    if iid:
        ins.training_samples.compute_importance.assert_not_called()
        ins.iid_samples.compute_importance.assert_called_once()
    else:
        ins.training_samples.compute_importance.assert_called_once()
        ins.iid_samples.compute_importance.assert_not_called()


def test_update_evidence(ins, iid):
    ins.draw_iid_live = iid
    ins.training_samples = MagicMock(spec=OrderedSamples)
    if iid:
        ins.iid_samples = MagicMock(spec=OrderedSamples)
    INS.update_evidence(ins)
    ins.training_samples.update_evidence.assert_called_once()
    if iid:
        ins.iid_samples.update_evidence.assert_called_once()


def test_update_sample_counts(ins):
    ins.samples_unit = {"it": np.array([-1, 0, 2, 2, 2])}
    ins.proposal = MagicMock()
    ins.proposal.n_proposals = 5
    INS.update_sample_counts(ins)
    assert ins.sample_counts == {-1: 1, 0: 1, 1: 0, 2: 3, 3: 0}



# -*- coding: utf-8 -*-
"""Tests related to drawing new points from the pool."""
import numpy as np
import pytest
from unittest.mock import MagicMock

from nessai.proposal import FlowProposal


def test_draw_populated(proposal):
    """Test the draw method if the proposal is already populated"""
    proposal.populated = True
    proposal.samples = np.arange(3)
    proposal.indices = list(range(3))
    out = FlowProposal.draw(proposal, None)
    assert out == proposal.samples[2]
    assert proposal.indices == [0, 1]


def test_draw_populated_last_sample(proposal):
    """Test the draw method if the proposal is already populated but there
    is only one sample left.
    """
    proposal.populated = True
    proposal.samples = np.arange(3)
    proposal.indices = [0]
    out = FlowProposal.draw(proposal, None)
    assert out == proposal.samples[0]
    assert proposal.indices == []
    assert proposal.populated is False


@pytest.mark.parametrize("update", [False, True])
def test_draw_not_populated(proposal, update, wait):
    """Test the draw method when the proposal is not populated"""
    proposal.populated = False
    proposal.poolsize = 100
    proposal.samples = None
    proposal.indices = []
    proposal.update_poolsize = update
    proposal.update_poolsize_scale = MagicMock()
    proposal.ns_acceptance = 0.5

    def mock_populate(*args, **kwargs):
        wait()
        proposal.populated = True
        proposal.samples = np.arange(3)
        proposal.indices = list(range(3))

    proposal.populate = MagicMock(side_effect=mock_populate)

    out = FlowProposal.draw(proposal, 1.0)

    assert out == 2
    assert proposal.populated is True

    proposal.populate.assert_called_once_with(1.0, N=100)

    if update:
        proposal.update_poolsize_scale.assert_called_once()
    else:
        proposal.update_poolsize_scale.assert_not_called()



# -*- coding: utf-8 -*-
"""
Base object for all proposal classes.
"""
from abc import ABC, abstractmethod
import datetime
import logging
import numpy as np

logger = logging.getLogger(__name__)


class Proposal(ABC):
    """
    Base proposal object

    Parameters
    ----------
    model: obj
        User-defined model
    """

    def __init__(self, model):
        self.model = model
        self.populated = True
        self._initialised = False
        self.training_count = 0
        self.population_acceptance = None
        self.population_time = datetime.timedelta()
        self.r = np.nan
        self.samples = []
        self.indices = []
        self._checked_population = True

    @property
    def initialised(self):
        """Boolean that indicates if the proposal is initialised or not."""
        return self._initialised

    @initialised.setter
    def initialised(self, boolean):
        """Setter for initialised"""
        if boolean:
            self._initialised = boolean
        else:
            self._initialised = boolean

    def initialise(self):
        """
        Initialise the proposal
        """
        self.initialised = True

    def evaluate_likelihoods(self):
        """Evaluate the likelihoods for the pool of live points."""
        self.samples["logL"] = self.model.batch_evaluate_log_likelihood(
            self.samples
        )

    @abstractmethod
    def draw(self, old_param):
        """
        New a new point given the old point
        """
        raise NotImplementedError

    def train(self, x, **kwargs):
        """
        Train the proposal method

        Parameters
        ----------
        x: array_like
            Array of live points to use for training
        kwargs:
            Any of keyword arguments
        """
        logger.error("This proposal method cannot be trained")

    def resume(self, model):
        """
        Resume the proposal with the model
        """
        self.model = model

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["model"]
        return state
